#!/usr/bin/python
#encoding: utf8
"""
This file contains functions calculating a few intermediate values, sanity checks, etc. in the tracking table
"""
import tables
import helper
import os
from error import warning, error

def parts((comp, ident), id, length, values, **args):
	"""calculate vehicle parts"""
	length = (lambda y: (8, warning("%s %s: Has no length set"%(comp,ident)))[0] if y=='' or int(y)==0
			else (16, warning("%s %s: Has too high length set"%(comp,ident)))[0] if int(y)>16
			else (3, warning("%s %s: Has too low length set"%(comp,ident)))[0] if int(y)<3
			else int(y))(length)
	slices = []
	if len(tables.parts[length]) > 1:
		# front = id+0x2000
		# middle = id+0x1000
		# end = id+0x0000
		id_mod = 0x2000
		for sl_length in tables.parts[length]:
			assert(id_mod >= 0)
			slices.append((id+id_mod, sl_length))
			id_mod -= 0x1000
		sl_pos = 1
	else:
		slices.append((id+0x3000, length))
		sl_pos = 0
	values["length"] = length
	values["slices"] = slices
	values["parts"] = len(slices)
	values["sl_pos"] = sl_pos

def artic_props((comp, ident), artic, parts, slices, sl_pos, artic_ident, artic_pattern, weight, power, max_te, cost_factor, run_cost_factor, values, traction_type, **args):
	head_traction_type = traction_type # use to ensure the middle running cost is correct	
	articulation = []
	if (artic or parts > 1) and not artic_ident == 'custom':
		offset = 0
		for (sl_id, sl_length) in slices:
			articulation.append(sl_id)
			offset += 1
		if artic:
			if artic_pattern == "":
				warning("%s %s: Has invalid articulated parts set"%(comp, ident))
			num_artic_parts = len(artic_pattern)
			cur_artic_part = 0 # Which part the loop is at

			while cur_artic_part < num_artic_parts:
				# artic_pattern is something like "3m,e" (would be three identical middle parts and one end part), commas are optional
				remaining_identical_artic_parts = 1 # = number articulated parts of the same type to append

				if artic_pattern[cur_artic_part] == ",":
					cur_artic_part += 1
					continue # Allow the use of commas as separators, but don't require it
				if artic_pattern[cur_artic_part].isdigit():
					# Allow the use of single digits to abbreviate long runs of identical letters
					remaining_identical_artic_parts = int(artic_pattern[cur_artic_part])					
					cur_artic_part += 1

				while remaining_identical_artic_parts > 0:
					if (comp, artic_ident+artic_pattern[cur_artic_part]) in tables.table:
						artic_values = tables.table[(comp, artic_ident+artic_pattern[cur_artic_part])]
					else:
						artic_values = values
					for (sl_id, sl_length) in artic_values["slices"]:
						articulation.append(sl_id)
						offset += 1
					weight += artic_values["weight"]
					power += artic_values["power"]
					max_te += artic_values["max_te"]
	
 					# Perform corrective action on "fake" MUs with an unpowered middle and/or end:

					if "Wagon" in artic_values["traction_type"]:
						# Adjust for the much smaller value of the wagon purchase cost base
						cost_factor += artic_values["cost_factor"] * 350 / 17580
					else:
						cost_factor += artic_values["cost_factor"]

					# Wagons use the steam running cost base	
					if "Diesel" in head_traction_type and "Wagon" in artic_values["traction_type"]:
						run_cost_factor += artic_values["run_cost_factor"] * 11200 / 15600	
				
					elif "Electric" in head_traction_type and "Wagon" in artic_values["traction_type"]:
						run_cost_factor += artic_values["run_cost_factor"] * 11200 / 19200

					else:
						run_cost_factor += artic_values["run_cost_factor"] 
					remaining_identical_artic_parts -= 1
				cur_artic_part += 1
	values["articulation"] = articulation
	values["weight"] = weight
	values["power"] = power
	values["max_te"] = max_te
	values["cost_factor"] = min(cost_factor, 255)
	values["run_cost_factor"] = min(run_cost_factor, 255)

def capacity((comp, ident), capacity, usage, values, **args):
	"""calculate vehicle capacity"""
	capacity = (lambda y: (0,warning("%s %s: Has no capacity set"%(comp, ident)))[0] if y=='' else min(255, int(y)))(capacity)
	loading_speed = min(255, (lambda z: # TODO: modifier for doubledecker wagons
			(z+1)/2 if usage in ["commuter","local/commuter","commuter/express"] else
			(z+3)/4 if usage in ["express","express/passenger","local/express","passenger,express"] else
			(z+2)/3
		)(capacity))
	if capacity not in tables.caps:
		tables.caps.append(capacity)
	if loading_speed not in tables.caps:
		tables.caps.append(loading_speed)
	values["capacity"] = capacity
	values["loading_speed"] = loading_speed

def refits(key, cargo_classes, values, **args):
	"""calculate vehicle refits"""
	refits = cargo_classes.split(',')
	if len(refits) == 1 and (refits[0] == '') or (refits[0] == '0'):
		values[   "allowed"] = "ALL_NORMAL_CARGO_CLASSES"
		values[ "forbidden"] = "NO_CARGO_CLASS"
		values["add_cargos"] = "[]"
		values["rem_cargos"] = "[]"
		return
	skip_allowed = False
	skip_forbidden = False
	allowed = []
	forbidden = []
	add_cargos = []
	rem_cargos = []
	for item in refits:
		remove = False
		if item.startswith('-'):
			remove = True
			item = item[1:]
		if item.startswith('+'):
			item = item[1:]
		if item == "none":
			if remove:
				values[  "forbidden"] = "NO_CARGO_CLASS"
				skip_forbidden = True
			else:
				values[  "allowed"] = "NO_CARGO_CLASS"
				skip_allowed = True
		elif item == "all":
			if remove:
				values[  "forbidden"] = "ALL_NORMAL_CARGO_CLASSES"
				skip_forbidden = True
			else:
				values[  "allowed"] = "ALL_NORMAL_CARGO_CLASSES"
				skip_allowed = True
		elif item in tables.cargo_classes:
			if remove:
				forbidden.append(tables.cargo_classes[item])
			else:
				allowed.append(tables.cargo_classes[item])
		else:
			if remove:
				rem_cargos.append(item)
			else:
				add_cargos.append(item)
	if not skip_allowed:
		values["allowed"] = " bitmask(" + ", ".join(allowed) + ")" # space is here because of sort order (see below)
	if not skip_forbidden:
		values["forbidden"] = "bitmask(" + ", ".join(forbidden) + ")"
	values["add_cargos"] = "[" + ", ".join(add_cargos) + "]"
	values["rem_cargos"] = "[" + ", ".join(rem_cargos) + "]"

def lifetime((comp, ident), avail, intro_date, last_built, decommissioned, values, **args):
	"""
	vehicle life == decommissioned - last built
	model life == decommissioned - first built
	availability == first built .. last built
	"""
	if len(avail) == 0:
		values["intro_date"]=0
		values["vehicle_life"]=0
		values["model_life"]=0
		values["retire_early"]=0
		return
	if intro_date == "":
		warning("%s %s: Has no intro date set"%(comp,ident))
		intro_date = 1920
	else:
		intro_date = int(intro_date)
	if last_built == "n/a":
		last_built = 2013
	elif last_built == "":
		warning("%s %s: Has no last built date set"%(comp,ident))
		last_built = intro_date+10
	else:
		last_built = min(map(int,last_built.split("/")))
	if last_built - intro_date < 8:
		last_built = intro_date+8
	if decommissioned == "n/a":
		decommissioned = last_built + 30
	elif decommissioned == "":
		warning("%s %s: Has no decommissioned date set"%(comp,ident))
		decommissioned = last_built + 20
	else:
		new_date = 0
		for date in decommissioned.split("/"):
			if date == "": continue
			new_date = max(new_date, int(date))
		decommissioned = new_date
		if decommissioned == 0:
			warning("%s %s: Has no decommissioned date set"%(comp,ident))
			decommissioned = last_built + 20
	values["intro_date"]=intro_date
	values["vehicle_life"]=max(20, decommissioned-last_built)
	values["model_life"]=decommissioned-intro_date
	values["retire_early"]=values["vehicle_life"]-4

def conditional_lifetime((comp, ident), obsoleted_by, avail, intro_date, model_life, retire_early, values, **args):
	"""
	extend or shorten model life, depending on availability of "obsoleted by" vehicles
	"""
	not_avail = []
	retire = {}
	retire_date = intro_date + model_life - retire_early - 4
	values["not_avail"] = not_avail
	values["retire"] = retire
	if len(avail) == 0:
		return
	if obsoleted_by == '':
		warning("%s %s: Has no obsoleted by set"%(comp, ident))
		return
	if obsoleted_by == 'none':
		return
	if obsoleted_by == 'inf':
		values["model_life"] = model_life + 60
		return
	for item in obsoleted_by.split(','):
		item = item.split(':')
		if len(item) == 1:
			comp2 = comp
			ident2 = item[0]
		else:
			comp2 = item[0]
			ident2 = item[1]
		try:
			values2 = tables.table[(comp2, ident2)]
		except KeyError:
			warning("%s %s: Has invalid obsoleted by: %s %s"%(comp, ident, comp2, ident2))
			continue
		intro_date2 = values2["intro_date"]
		if intro_date2 < intro_date:
			for avail2 in values2["avail"]:
				if avail2 not in avail and avail2 not in not_avail:
					not_avail.append(avail2)
		else:
			for avail2 in values2["avail"]:
				if avail2 not in retire:
					retire[avail2] = intro_date2-retire_date
				else:
					retire[avail2] = min(intro_date2-retire_date, retire[avail2])
	if retire:
		values["model_life"] = "%s + (%s:0)"%(model_life, ':'.join(map(lambda (key, value): "%s?%s"%(key, value), sorted(retire.iteritems(), cmp=lambda (key1, value1), (key2, value2): cmp(value1, value2)))))

def availability(avails, (comp, ident), company, epoch, set, values, **args):
	try:
		values["avail"] = helper.make_avail(avails, company.split(','), epoch.split(','), set.split(','))
	except TypeError:
		warning("%s %s: Has invalid availability set: %s %s %s"%(comp, ident, company, epoch, set))
		values["avail"] = helper.make_avail(avails, (comp,), (epoch.split(',')[0],), (set.split(',')[0],))

def availability2(key, avail, not_avail, values, **args):
	if avail:
		if not_avail:
			values["availability"] = "(%s) && !(%s)"%(' || '.join(map(str, avail)), ' || '.join(map(str, not_avail)))
		else:
			values["availability"] = "%s"%(' || '.join(map(str, avail)),)
	else:
		values["availability"] = "0"

def names((comp, ident), long_name, translatable_name, values, **args):
	"""preprocess the vehicle names"""
	r = helper.sup_regexp
	m = r.match(long_name)
	while m is not None:
		long_name = m.group(1)+helper.replace_sup(m.group(2))+m.group(3)
		m = r.match(long_name)
	if translatable_name:
		#validate the number of {STRING} tags in the name
		translatable_name = map(lambda s: "string(STR_NAME_ENG_%s)"%(s,),translatable_name.split(','))
		n = len(translatable_name)
		r = helper.string_regexp
		m = r.match(long_name)
		tmp_name = long_name
		i = 0
		while m is not None:
			tmp_name = m.group(1)+m.group(2)
			m = r.match(tmp_name)
			i+=1
		if n < i:
			warning("%s %s: has not enough string parameters: %s found, %s expected"%(comp, ident, n, i))
			for j in range(n, i):
				translatable_name.append("string(STR_EMPTY)")
		if n > i:
			warning("%s %s: has too many string parameters: %s found, %s expected"%(comp, ident, n, i))
			translatable_name = translatable_name[:i]
		values["translatable_name"] = ','.join(translatable_name)
	values["long_name"] = long_name

def strings(auto_strings, key, axle_scheme, values, **args):
	"""generate list of strings"""
	if axle_scheme not in auto_strings:
		auto_strings.append(axle_scheme)
	values["axle_scheme_string"] = "STR_AUTO_%d"%(auto_strings.index(axle_scheme),)

def recolour(key, recolour, values, **args):
	"""replace "generic" recolour map"""
	if recolour == "generic":
		values["recolour"] = "cargo:(GRAI:grai,SULP:sulp,SAND:sand,AORE:aore,CORE:core,FERT:fert,IORE:iore,CLAY:clay,OLSD:olsd,LIME:lime,POTA:pota,GRVL:grvl,COAL:coal,FRUT:frut,SCMT:scmt,blue)"

def misc(usage_strings, fixed_strings, (comp, ident), artic_ident, max_speed, weight, power, max_te, cost_factor, run_cost_factor, usage, values, **args):
	"""calculate some miscellaneous values"""
	values["weight"] = (lambda y: (100, warning("%s %s: Has no weight set"%(comp,ident)))[0] if y=='' else float(y))(weight)
	values["cost_factor"] = (lambda y: (1, warning("%s %s: Has no cost set"%(comp,ident)))[0] if y=='' else max(1,min(float(y),255)))(cost_factor)
	values["run_cost_factor"] = (lambda y: (1, warning("%s %s: Has no cost set"%(comp,ident)))[0] if y=='' else max(1,min(float(y),255)))(run_cost_factor)
	values["power"] = float((lambda y: (1000, warning("%s %s: Has no power set"%(comp,ident)))[0] if y=='' else y)(power))
	values["max_te"] = float((lambda y: (100, warning("%s %s: Has no tractive effort set"%(comp,ident)))[0] if y=='' else y)(max_te))
	values["max_speed"] = (lambda y: (255, warning("%s %s: Has no speed set"%(comp,ident)))[0] if y=='' or '/' in y else int(y))(max_speed)
	values["artic"] = artic_ident not in ("none", "custom")
	values["usage_string"] = usage_strings[usage][0]
	fixed_strings.add(values["usage_string"])
	values["ai_special_flag"] = usage_strings[usage][1]

def track_type(fixed_strings, key, axle_weight, max_speed, traction_type, values, **args):
	axle_weight = (lambda y: (0, warning("%s %s: Has no axle weight set"%(comp,ident)))[0] if y=='' else float(y))(axle_weight)
	track_type = (
		# traction type
		(lambda x: "N" if x in ["Steam", "Diesel", "Battery", "Wagon"] else "E" if x in ["Electric"] else "3" if x == "3rd Rail" else ("N", warning("%s %s: Has invalid traction type set"%(comp,ident)))[0])(traction_type.split("/")[0]),
		# track class
		"A" if axle_weight <= 16.8 else "B" if axle_weight <= 18.9 else "C" if axle_weight < 21.0 else "D",
		# speed limit
		"V" if max_speed <= 80 else "L" if max_speed <= 120 else "M" if max_speed <= 180 else "H",
		# narrow gauge
		"n" if "Narrow" in traction_type.split("/") else "",
		)
	values["track_type_classes"] = 'tracktype_classes%s%s%s'%(track_type[1], track_type[0], track_type[3])
	values["track_type_speed"]   = 'tracktype_speed%s%s%s'  %(track_type[2], track_type[0], track_type[3])
	values["track_type_default"] = 'tracktype_default%s%s'  %(track_type[0], track_type[3])
	values["track_class_string"] = 'STR_TRACK_CLASSES_%s'%(track_type[1],)
	values["track_speed_string"] = 'STR_TRACK_SPEED_%s'  %(track_type[2],)
	values["track_type_string"]  = 'STR_TRACK_TYPE_%s%s' %(track_type[0], track_type[3])
	fixed_strings.add(values["track_class_string"])
	fixed_strings.add(values["track_speed_string"])
	fixed_strings.add(values["track_type_string"])

def unique((comp, ident), id, values, **args):
	"""check uniqueness of IDs"""
	if id is None:
		return
	if tables.ids[id]:
		error('ID# %d is duplicate between %s %s and %s %s'%(id, tables.ids[id][0], tables.ids[id][1], comp, ident))
	tables.ids[id]=(comp,ident)

def unused():
	unused = []
	first = 0
	last = 0
	for n,a in enumerate(tables.ids):
		if a is None:
			if last < n-1:
				first = n
			last = n
		else:
			if last != 0 and last == n-1:
				unused.append((first, last))

	if unused:
		s = "unused IDs: "
		for first,last in unused:
			if first == last:
				s+='%d '%(last,)
			elif first+1 == last:
				s+='%d %d '%(first, last)
			else:
				s+='%d..%d '%(first, last)
		warning(s)
